theory Basic_Refinements
imports Explore_Tree
begin

  subsection \<open>Part I: Refinement to efficient data structure\<close>    

  text \<open>
    Recall the workset algorithm from @{theory Explore_Tree}.
    We will now refine the workset to a (distinct) list,
    and the graph to a successor function, returning a distinct 
    list of elements. 

    The Isabelle Collections Framework provides some basic collection 
    data structures, including distinct lists (based on @{typ "'a dlist"}) to implement sets.

    The correctness theorems for the basic operations are 
    \<close>
  thm ls.correct

  text \<open>Selecting an element from a list\<close>
  term ls.sel
  find_theorems ls.sel

  text \<open> Note, that the ICF provides some more data structures,
    e.g.
    \<close>
  thm rs.correct rm.correct -- \<open>Red black tree based sets and maps\<close>
  thm hs.correct hm.correct -- \<open>Hash sets/maps based on RBT\<close>
  thm ts.correct tm.correct -- \<open>Tries\<close>
  thm ahs.correct ahm.correct -- \<open>Hash sets/maps based on array (efficient if used linearly)\<close>
  thm iam.correct ias.correct -- \<open>Sets/maps of nat, based on array (efficient if used linearly)\<close>
  thm alprioi.correct aluprioi.correct -- \<open>Priority queues/maps based on Finger Trees\<close>
  thm skew.correct -- \<open>Priority queues based on skew heaps\<close>

  thm lsi.correct lmi.correct -- \<open>Distinct lists, with explicit invariant, based on @{typ "'a list"}\<close>

  thm lsnd.correct -- \<open>List-set by non-distinct lists\<close>
  thm lss.correct -- \<open>List-set by sorted lists\<close>

  text \<open>This part of the ICF is the old ICF, which is intended to be used manually.
    Apart from concrete implementations, it offers a mechanism to specify the
    interfaces of the data structures (You have probably noted that all data structures
    for sets (maps,...) have the same operations), and a library of generic algorithms,
    which are parameterized over the interfaces. However, this tutorial 
    will not cover the ICF in detail. We refer you to the ICF user guide,
    located at @{text "Collections/Userguides/ICF_Userguide"}.

    The old ICF has some restrictions, for example, it allows no refinement of the elements of
    a collection data structure.
    The new ICF is based on the Autoref-Tool, and will be covered 
    later in this tutorial, when introducing the Autoref tool.
    \<close>


  text \<open> Back to the initial problem: We want to refine the graph to a successor function,  
    which returns a list of nodes.

    Using distinct lists from the ICF to represent the set of successors,
    we get the following refinement relation:
  \<close>  
  definition "succ_\<alpha> succ \<equiv> { (u,v) . v \<in> ls.\<alpha> (succ u) }"
  definition "succ_invar succ \<equiv> \<forall>u. ls.invar (succ u)"
  abbreviation "succ_rel \<equiv> br succ_\<alpha> succ_invar"

  text \<open>Note that the invariant is vacuous, as @{term ls.invar} is. 
    We only include it for completeness here\<close>
  lemma [simp]: "succ_invar \<equiv> \<lambda>_. True"
    unfolding succ_invar_def[abs_def] by auto

  (*definition "succ_edges_rel \<equiv> { (succ,E). \<forall>v. (succ v, E``{v}) \<in> \<langle>Id\<rangle>ls.rel }"*)

  text \<open> Formally, the parameter to the relator @{text ls.rel} is the refinement relation
    for the elements. However, due to restrictions of the old ICF, which is used here,
    this must always be @{term Id}. \<close>

  text \<open> We have one interesting operation on graphs, namely returning the set of successors.
    We have to show its correctness as follows:
    \<close>
  lemma succ_correct: 
    "succ_invar succ \<Longrightarrow> ls.\<alpha> (succ v) = (succ_\<alpha> succ)``{v}"
    "succ_invar succ \<Longrightarrow> ls.invar (succ v)"  -- \<open>Note, vacuous statement, as ls has no invariant\<close>
    unfolding succ_\<alpha>_def by (auto simp: br_def)

  definition "impl_ws succ v0 vt \<equiv> do {
    let s = (ls.sng v0,False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILE (\<lambda>(ws,r). \<not>r \<and> \<not>ls.isEmpty ws) (\<lambda>(ws,_). do {
      let v = the (ls.sel ws (\<lambda>_. True));  (* Pick element from workset *)
      let ws = ls.delete v ws;     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ls.union ws (succ v);  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
  }"

  lemma impl_ws_refine:
    assumes A: "T = succ_\<alpha> succ"
    shows "impl_ws succ v0 vt \<le> \<Down>Id (abs_ws T v0 vt)"
    unfolding impl_ws_def abs_ws_def A
    apply refine_rcg      -- \<open>Invoke VCG\<close>
    apply refine_dref_type -- \<open>Heuristics to instantiate refinement relations\<close>
    apply (auto simp add: br_def ls.correct ls.pick_ref)
    apply (auto simp add: succ_correct)
      -- \<open>The VCs go away almost automatically. There is some interference 
        between @{thm [source] ls.pick_ref} and @{thm [source] succ_correct}\<close>
    done

  text \<open>The heuristics used by @{text refine_dref_type} is 
    a simple, rule-based heuristics, that uses the abstract and concrete type
    to infer an appropriate relation\<close>  
  thm refine_dref_RELATES -- \<open>Rules to infer relation and relators\<close>
  thm refine_dref_pattern -- \<open>Patterns in refinement VCs that are processed\<close>

  text \<open>Note: Refinement relations may also be instantiated manually\<close>  
  lemma 
    assumes A: "T = succ_\<alpha> succ"
    shows "impl_ws succ v0 vt \<le> \<Down>Id (abs_ws T v0 vt)"
    unfolding impl_ws_def abs_ws_def A
    apply refine_rcg      -- \<open>Invoke VCG\<close>
    apply (rule introR[where R="\<langle>Id\<rangle>ls.rel \<times>\<^sub>r bool_rel"])
    apply (auto simp add: br_def ls.correct ls.pick_ref)
    apply (auto simp add: succ_correct)
    done

  subsection \<open>Part II: Transfer to deterministic monad\<close>    

  text \<open>The last step towards an executable specification is to get rid 
    of the nondeterminism monad, which is not executable (at least not efficiently).
    
    A deterministic program can be transferred to a deterministic monad,
    preserving its structure. This transfer is canonical, and is automated
    using the @{text refine_transfer} tactic.

    The deterministic monad has type @{typ "'a dres"}:
  \<close>
  term dRETURN term dFAIL term dSUCCEED
  thm dres_monad_laws

  text \<open>The relation to the nondeterministic monad is established via @{const nres_of},
    and a set of rules to transfer the combinators.\<close>
  term nres_of
  thm det_bind det_RETURN dres.transfer_RECT
  thm refine_transfer -- \<open>All the rules, and some more for plain-function synthesis (see below)\<close>

  text \<open>  
    Note that Isabelle/HOL has a very powerful built-in synthesis mechanism:
    A schematic variable can be instantiated by resolution with rules. 
    The @{text schematic_goal} statement allows to explicitly include
    schematic variables into a lemma statement. The proven lemma will have the
    schematic instantiated.

    Thus, the following lemma tries to synthesize a program @{text "?c"}
    in the deterministic monad, which refines the implementation.
    \<close>
  schematic_goal code_ws: "nres_of ?c \<le> impl_ws succ v0 vt"
    unfolding impl_ws_def
    apply (refine_transfer (post))
    done
  text \<open> The @{text concrete_definition} command defines constants from theorems.
    It provides a new theorem with the constant folded. 
    The optional for-clause determines the order of the parameters \<close>  
  concrete_definition code_ws for succ ?v0.0 vt uses code_ws
  print_theorems
  thm code_ws.refine
  text \<open>Note: The @{text concrete_definition} command has a set of default 
    patterns to determine which part of the theorem is extracted. However, 
    the patterns are fully configurable.\<close>

  text \<open>The next command converts recursion combinators to equations,
    and sets up the code generator accordingly.
    This is required because the code generator cannot cope with recursion 
    combinators, but requires equations. Note: In this case, we have
    no recursion combinators, so the command effectively does nothing.\<close>
  prepare_code_thms code_ws_def
  print_theorems
  text \<open>The deterministic monad is executable. Also, the data structures
    from the collection framework are executable, such that we can generate 
    code.\<close>
  export_code code_ws in SML
  text \<open>Note that the generated code contains some unused module definitions.
    These are the setups of various libraries that come with CAVA,
    on which these theories are based. 
    Unfortunately, Isabelle cannot include them only on demand.

    For now, you may just ignore them.
    \<close>

  text \<open>Finally, we combine the refinement theorems, to
    get a direct correctness statement for our final algorithm\<close>
  lemma code_ws_correct:
    fixes succ
    defines "T \<equiv> succ_\<alpha> succ"
    shows "case code_ws succ v0 vt of 
      dSUCCEEDi \<Rightarrow> True              (* May not terminate *)
    | dRETURN r \<Rightarrow> r \<longrightarrow> (v0,vt)\<in>T\<^sup>*  (* Return value satisfies spec *)
    | dFAILi \<Rightarrow> False                (* Will not fail *)
    "
  proof -  
    txt \<open>This easily follows by transitivity. 
      Note that the transitivity reasoner is set up to also
      cope with data refinements (However, this feature is 
      not required here)\<close>
    note code_ws.refine[of succ v0 vt]
    also note impl_ws_refine[where T=T]
    also note abs_ws_correct
    finally show ?thesis 
      using assms unfolding T_def 
      by (auto split: dres.splits)
  qed

  text \<open>Let's play around with the generated code a bit. 
    For better handling, we restrict the type to integers\<close>
  definition "code_ws_nat \<equiv> code_ws :: _ \<Rightarrow> nat \<Rightarrow> _"
  ML_val \<open>
    val code_ws = @{code code_ws_nat}
    fun succ 0 = [1] | succ 1 = [2,3] 
      | succ 10 = [11] | succ 11 = [10]
      | succ _ = []

    val nat_of_integer = @{code nat_of_integer}
    val integer_of_nat = @{code integer_of_nat}
    fun succ' n = @{code Dlist} (map nat_of_integer (succ (integer_of_nat n)))

    fun path u v = @{code code_ws_nat} 
      succ' (nat_of_integer u) (nat_of_integer v)

    val res = (path 0 2,path 1 2, path 2 1, path 2 3)
    (*val res = (path 10 2) (* does not terminate!*) *)
    \<close>


  subsubsection \<open>Refinement to plain function\<close>
  text \<open> For partial correct programs, the correctness result will
    always include the possibility of nontermination.

    However, for total correct programs, we can exclude this possibility,
    and synthesize a function that returns a plain value, instead of
    an @{typ "'a dres"}. 

    The simplest way of doing so is to prove that the generated code
    cannot have value @{term dSUCCEED}, and extract the result by a selector.
    \<close>
  thm the_resI

  text \<open>So lets define a total correct version of the worklist algorithm.
    Note, we will prove no correctness lemmas here. \<close>

  definition "impl_wsT succ v0 vt \<equiv> do {
    let s = (ls.sng v0,False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILET (\<lambda>(ws,r). \<not>r \<and> \<not>ls.isEmpty ws) (\<lambda>(ws,_). do {
      let v = the (ls.sel ws (\<lambda>_. True));  (* Pick element from workset *)
      let ws = ls.delete v ws;     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ls.union ws (succ v);  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
  }"

  text \<open>@{text refine_transfer} contains theorems to prove 
    @{text "_ \<noteq> dSUCCEED"} automatically for all total-correct 
    combinators.\<close>
  schematic_goal "RETURN (the_res ?c) \<le> impl_wsT succ v0 vt"
    unfolding impl_wsT_def
    apply (refine_transfer (post) the_resI)
    done

  text \<open>However, if the only recursion through the monad is tail-recursion 
    (while, foreach), we can even get rid of the monad alltogether.
    This is because HOL allows the definition of tail-recursive functions
    without a monad, even if no termination proof can be given.
  \<close>  
  thm dres_while.transfer_WHILET_esc

  schematic_goal "RETURN ?c \<le> impl_wsT succ v0 vt"
    unfolding impl_wsT_def
    apply (refine_transfer (post))
    done


  subsubsection \<open>Handling Arbitrary Recursion\<close>
  text \<open>To illustrate the slight differences if arbitrary recursion is present, 
    let's refine the recursive DFS function:\<close>
  definition "find_node_impl RECcomb succ v0 vt \<equiv> RECcomb (\<lambda>rec v. 
    if v=vt then   (* Return true if node is found *)
      RETURN True
    else           (* Otherwise, iterate over successors *)
      FOREACHc  
        (ls.\<alpha> (succ v))    (* Successors of node v *) 
        (\<lambda>x. \<not>x)    (* Iterate until result is true *)
        (\<lambda>v _. rec v) (* Invoke recursively *)
        False       (* Initial state for loop *) 
  ) v0"

  abbreviation "find_node_total_impl \<equiv> find_node_impl RECT"
  
  lemma find_node_total_impl_refine:
    assumes A: "T = succ_\<alpha> succ"
    shows "find_node_total_impl succ v0 vt \<le> \<Down>Id (find_node_total T v0 vt)"
    unfolding find_node_impl_def find_node_def A
    apply (refine_rcg inj_on_id) -- \<open>The foreach-rules 
      allow data refinement on the iterated set. Using @{thm [source] inj_on_id}
      is the standard way to instantiate this data refinement to identity.\<close>
    apply refine_dref_type
    apply (auto simp: succ_correct)
    done

  schematic_goal find_node_total_code: "RETURN ?c \<le> find_node_total_impl succ v0 vt"  
    unfolding find_node_impl_def
    apply (refine_transfer the_resI) -- \<open>You cannot refine arbitrary 
      recursion to plain functions, thus, you have to use the monad.
      However, you may extract a plain result from the computation, as
      we do here by using @{thm [source] the_resI}.\<close>
    done
  concrete_definition find_node_total_code for succ ?v0.0 vt uses find_node_total_code

  text \<open>Having recursion combinators, the @{text prepare_code_thms} 
    command is mandatory:\<close>
  prepare_code_thms find_node_total_code_def
  text \<open>Inspecting the produces theorems, one sees how recursion was unfolded
    into equations:\<close>
  print_theorems

  export_code find_node_total_code in SML


end

