theory Basic_RefinementsT
imports Explore_TreeT
begin

  subsection \<open>Part I: Refinement to efficient data structure\<close>    

  text \<open>
    Recall the workset algorithm from @{theory Explore_TreeT}.
    We will now refine the workset to a (distinct) list,
    and the graph to a successor function, returning a distinct 
    list of elements. 

    The Isabelle Collections Framework provides some basic collection 
    data structures, including distinct lists (based on @{typ "'a dlist"}) to implement sets.

    The correctness theorems for the basic operations are 
    \<close>
  thm ls.correct

  text \<open>Selecting an element from a list\<close>
  term ls.sel
  find_theorems ls.sel

  text \<open> Note, that the ICF provides some more data structures,
    e.g.
    \<close>
  thm rs.correct rm.correct -- \<open>Red black tree based sets and maps\<close>
  thm hs.correct hm.correct -- \<open>Hash sets/maps based on RBT\<close>
  thm ts.correct tm.correct -- \<open>Tries\<close>
  thm ahs.correct ahm.correct -- \<open>Hash sets/maps based on array (efficient if used linearly)\<close>
  thm iam.correct ias.correct -- \<open>Sets/maps of nat, based on array (efficient if used linearly)\<close>
  thm alprioi.correct aluprioi.correct -- \<open>Priority queues/maps based on Finger Trees\<close>
  thm skew.correct -- \<open>Priority queues based on skew heaps\<close>

  thm lsi.correct lmi.correct -- \<open>Distinct lists, with explicit invariant, based on @{typ "'a list"}\<close>

  thm lsnd.correct -- \<open>List-set by non-distinct lists\<close>
  thm lss.correct -- \<open>List-set by sorted lists\<close>

  text \<open>This part of the ICF is the old ICF, which is intended to be used manually.
    Apart from concrete implementations, it offers a mechanism to specify the
    interfaces of the data structures (You have probably noted that all data structures
    for sets (maps,...) have the same operations), and a library of generic algorithms,
    which are parameterized over the interfaces. However, this tutorial 
    will not cover the ICF in detail. We refer you to the ICF user guide,
    located at @{text "Collections/Userguides/ICF_Userguide"}.

    The old ICF has some restrictions, for example, it allows no refinement of the elements of
    a collection data structure.
    The new ICF is based on the Autoref-Tool, and will be covered 
    later in this tutorial, when introducing the Autoref tool.
    \<close>


  text \<open> Back to the initial problem: We want to refine the graph to a successor function,  
    which returns a list of nodes.

    Using distinct lists from the ICF to represent the set of successors,
    we get the following refinement relation:
  \<close>  

  (* succ_\<alpha>, succ_invar, succ_rel*)
  definition "succ_\<alpha> succ = { (u,v). v \<in> ls.\<alpha> (succ u) }"
  definition "succ_invar succ \<equiv> \<forall>u. ls.invar (succ u)"
  abbreviation "succ_rel \<equiv> br succ_\<alpha> succ_invar" 
  lemma [simp]: "succ_invar = (\<lambda>_. True)"
    unfolding succ_invar_def[abs_def] by auto

  (* succ_correct *)
  lemma succ_correct: 
    "ls.\<alpha> (succ v) = (succ_\<alpha> succ)``{v}"
    "succ_invar succ \<Longrightarrow> ls.invar (succ v)"
    unfolding succ_\<alpha>_def by auto

  (* impl_ws*)

  definition "impl_ws succ v0 vt \<equiv> do {
    let s = (ls.sng v0,False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILE (\<lambda>(ws,r). \<not>r \<and> \<not>ls.isEmpty ws) (\<lambda>(ws,_). do {
      let v = the (ls.sel ws (\<lambda>_. True));  (* Pick element from workset *)
      let ws = ls.delete v ws;     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ls.union ws (succ v);  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
    }"

  (* impl_ws_refine *)
  lemma impl_ws_refine: 
    fixes succ
    defines "E \<equiv> succ_\<alpha> succ"
    shows "impl_ws succ v0 vt \<le> \<Down>Id (abs_ws E v0 vt)"
    unfolding impl_ws_def abs_ws_def
    apply refine_rcg
    apply refine_dref_type
    unfolding E_def
    apply (auto simp add: br_def ls.correct ls.pick_ref)
    apply (auto simp add: succ_correct)
    done

  text \<open>The heuristics used by @{text refine_dref_type} is 
    a simple, rule-based heuristics, that uses the abstract and concrete type
    to infer an appropriate relation\<close>  
  thm refine_dref_RELATES -- \<open>Rules to infer relation and relators\<close>
  thm refine_dref_pattern -- \<open>Patterns in refinement VCs that are processed\<close>

  text \<open>Note: Refinement relations may also be instantiated manually. \<close> 
  (* introR*)
  lemma 
    assumes A: "T = succ_\<alpha> succ"
    shows "impl_ws succ v0 vt \<le> \<Down>Id (abs_ws T v0 vt)"
    unfolding impl_ws_def abs_ws_def
    apply refine_rcg
    apply (rule introR[where R=" br ls.\<alpha> ls.invar \<times>\<^sub>r bool_rel"])
    oops

  subsection \<open>Part II: Transfer to deterministic monad\<close>    
  term dRETURN term dFAIL term dSUCCEED
  thm dres_monad_laws

  text \<open>The relation to the nondeterministic monad is established via @{const nres_of},
    and a set of rules to transfer the combinators.\<close>
  term nres_of
  thm det_bind det_RETURN dres.transfer_RECT
  thm refine_transfer -- \<open>All the rules, and some more for plain-function synthesis (see below)\<close>

  (* Show refine_transfer, concrete_definition, export_code, correctness lemma*)
  schematic_goal code_ws: "nres_of ?c \<le> impl_ws succ v0 vt"
    unfolding impl_ws_def
    by refine_transfer
  concrete_definition code_ws for succ ?v0.0 vt uses code_ws
  thm code_ws.refine

  export_code code_ws in SML

  definition "impl_wsT succ v0 vt \<equiv> do {
    let s = (ls.sng v0,False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILET (\<lambda>(ws,r). \<not>r \<and> \<not>ls.isEmpty ws) (\<lambda>(ws,_). do {
      let v = the (ls.sel ws (\<lambda>_. True));  (* Pick element from workset *)
      let ws = ls.delete v ws;     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ls.union ws (succ v);  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
    }"

  schematic_goal code_wsT: "RETURN ?c \<le> impl_wsT succ v0 vt"
    unfolding impl_wsT_def
    by (refine_transfer)
  concrete_definition code_wsT for succ ?v0.0 vt uses code_wsT
  thm code_wsT.refine

  export_code code_wsT in SML

  (* Play around with code in ML_val *)

  definition "code_wsT_nat \<equiv> (code_wsT :: _ \<Rightarrow> nat \<Rightarrow> _)"

  ML_val {*
    val ws = @{code code_wsT_nat}
    
    fun succ 0 = [1] | succ 1 = [2,3] 
      | succ 10 = [11] | succ 11 = [10]
      | succ _ = []
      
    fun csucc succ v = @{code Dlist} (map @{code nat_of_integer} (succ (@{code integer_of_nat} v)))  

    val v0 = @{code nat_of_integer} 0
    val vt = @{code nat_of_integer} 3

    val res = ws (csucc succ) v0 vt 


  *}

  (* To plain function. the_resI. (Need total correct version) *)


end

