theory Setsum
imports Tutorial_Basic
begin

text \<open>
  This theory gives a very short overview about code generation with 
  the refinement framework. It does not yet go into technical details.

  If you read this a a refinement framework beginner, you need not understand 
  any detail, but only the rough approach.
\<close>

text \<open>We may represent finite sets represented by distinct lists, 
  or by red-black trees, provided the elements are linearly ordered.
  
  Now, let's regard implementations of setsum, i.e., the sum of the elements 
  of a set.
\<close>

text \<open>
  For lists, we fold over the list, adding up the elements.
\<close>
lemma 
  assumes A: "distinct l"  
  shows "List.fold op + l 0 = \<Sum>(set l)"
proof -
  { fix a
    from A have "List.fold op + l a = \<Sum>(set l) + a"
      apply (induction l arbitrary: a)
      apply (auto simp: algebra_simps)
      done
  } thus ?thesis by simp
qed

text \<open>
  For red-black trees, we fold over the tree, adding up the elements.
\<close>
lemma 
  includes rbt.lifting
  fixes t :: "('k::{linorder,comm_monoid_add},'v) rbt"
  shows "RBT.fold (\<lambda>k _ r. r+k) t 0 = \<Sum>(set (RBT.keys t))"
proof transfer
  fix t :: "('k,'v) RBT_Impl.rbt"
  assume "is_rbt t" hence A: "rbt_sorted t" by simp

  { fix t1 t2 and a :: 'k
    assume "t1 |\<guillemotleft> a" "a \<guillemotleft>| t2"
    hence "set (RBT_Impl.keys t1) \<inter> set (RBT_Impl.keys t2) = {} 
      \<and> a\<notin>set (RBT_Impl.keys t1) \<and> a\<notin>set (RBT_Impl.keys t2)"
      unfolding rbt_less_prop rbt_greater_prop
      by force
  } note aux=this

  {
    fix a
    from A have "RBT_Impl.fold (\<lambda>k _ r. r+k) t a = \<Sum>(set (RBT_Impl.keys t)) + a"
      apply (induction t arbitrary: a)
      apply simp
      apply clarsimp
      apply (drule (1) aux)
      apply (auto simp: algebra_simps setsum_Un2 Diff_triv inf_commute) 
      done
  } thus "RBT_Impl.fold (\<lambda>k _ r. r + k) t 0 = \<Sum>(set (RBT_Impl.keys t))" 
    by simp
qed  

text \<open>
  We have now seen the same idea twice. 
  In both cases, we fold over the elements of the set, and add them up.
  However, we had to do the proof twice, both times coping with various
  technical details of the implementations.
\<close>

text \<open>
  Refinement allows us, to first prove the abstract algorithmic idea
  --- iterate over the set and sum up --- and then refine it to the
  various implementations.
\<close>

text \<open>
  Let's first define the abstract algorithm, using the 
  nondeterminism monad and the foreach combinator of
  the refinement framework.
\<close>
definition "abs_sum s \<equiv> do {
  FOREACH s (\<lambda>x r. RETURN (x+r)) 0
}"

text \<open>
  The correctness proof is fairly straightforward, using the
  verification condition generator, and hinting the 
  correct invariant for the foreach-loop.
\<close>
lemma abs_sum_correct:
  assumes "finite s"
  shows "abs_sum s \<le> SPEC (\<lambda>r. r = \<Sum>s)"
  using assms
  unfolding abs_sum_def apply -
  apply (refine_vcg FOREACH_rule[where I="\<lambda>it r. r = \<Sum>(s - it)"])
  apply (auto simp: it_step_insert_iff)
  done

text \<open>
  Next, let's define a version on lists
\<close>
definition "ls_sum l \<equiv> do {
  FOREACH (ls.\<alpha> l) (\<lambda>x r. RETURN (x+r)) 0
}"

text \<open>
  It's straightforward to show that this version refines the abstract
  algorithm:
\<close>
lemma ls_sum_refine: "(l,s)\<in>\<langle>Id\<rangle>ls.rel \<Longrightarrow> ls_sum l \<le> \<Down>Id (abs_sum s)"
  unfolding ls_sum_def abs_sum_def
  apply (refine_rcg inj_on_id)
  apply (auto simp: br_def)
  done

text \<open>
  And we can do exactly the same for red-black trees
\<close>
definition "rs_sum t \<equiv> do {
  FOREACH (rs.\<alpha> t) (\<lambda>x r. RETURN (x+r)) 0
}"

lemma rs_sum_refine: "(t,s)\<in>\<langle>Id\<rangle>rs.rel \<Longrightarrow> rs_sum t \<le> \<Down>Id (abs_sum s)"
  unfolding rs_sum_def abs_sum_def
  apply (refine_rcg inj_on_id)
  apply (auto simp: br_def)
  done

text \<open>Finally, to obtain executable code, we need to get out of the nondetermism monad\<close>
schematic_goal ls_sum_impl: "RETURN ?c \<le> ls_sum l"
  unfolding ls_sum_def
  by (refine_transfer (post))
concrete_definition ls_sum_impl uses ls_sum_impl
export_code ls_sum_impl in SML

text \<open>
  Note that @{text concrete_definition} extracts a pattern from a theorem, 
  defines it as a constant, and provides a theorem with the constant folded.

  It is set up to automatically recognize refinement patterns, but
  is fully customizable to your own patterns.
\<close>

notepad begin
  assume "(a::nat) < b"
  also assume "Q \<Longrightarrow> b < c"
  also assume "Q\<Longrightarrow>c \<le> d"
  finally have True by simp
end

text \<open>Combining the correctness and refinement lemmas (by transitivity) yields\<close>
lemma "(l,s)\<in>\<langle>Id\<rangle>ls.rel \<Longrightarrow> ls_sum_impl l = \<Sum>s"
proof -
  assume A: "(l,s)\<in>\<langle>Id\<rangle>ls.rel"
  hence B: "finite s" by (auto simp: br_def)
  note ls_sum_impl.refine
  also note ls_sum_refine[OF A]
  also note abs_sum_correct[of s, OF B]
  finally show ?thesis 
    by (auto)
qed  

text \<open>And the same for RBT\<close>
schematic_goal rs_sum_impl: "RETURN ?c \<le> rs_sum l"
  unfolding rs_sum_def
  by (refine_transfer (post))
concrete_definition rs_sum_impl uses rs_sum_impl
export_code rs_sum_impl in SML

lemma "(l,s)\<in>\<langle>Id\<rangle>rs.rel \<Longrightarrow> rs_sum_impl l = \<Sum>s"
proof -
  assume A: "(l,s)\<in>\<langle>Id\<rangle>rs.rel"
  note rs_sum_impl.refine
  also note rs_sum_refine
  also note abs_sum_correct[of s]
  finally show ?thesis using A 
    by (auto simp: br_def)
qed  

text \<open>Note, that most of this code was boilerplate, there were no original ideas
  but coming up with the correct abstract invariant. 

  In particular for bigger algorithms, where both, implementation and 
  abstract algorithm require some ideas, the separation of concerns
  between abstract and concrete proof makes up the effort of writing the 
  boilerplate code by far.

  Of course, much of this boilerplate could (and indeed can) be automated:
\<close>

text \<open>Lets start after the abstract algorithm again\<close>

schematic_goal ls_sum_impl':
  assumes [autoref_rules]: "(l,s)\<in>\<langle>nat_rel\<rangle>ls.rel"
  shows "(RETURN ?c,abs_sum s)\<in> \<langle>nat_rel\<rangle>nres_rel"
  unfolding abs_sum_def
  apply (autoref_monadic (plain))
  done
concrete_definition ls_sum_impl' uses ls_sum_impl'
export_code ls_sum_impl' in SML

text \<open>Here, we used the Autoref-tool, to automate 
  generation of the refined algorithm, the refinement proof,
  and the extraction from the monad. The concrete datatype
  to be used is determined by the @{text "[autoref_rules]"}-attributed
  assumption.

  Note that we restrict the type to natural numbers here. 
  A synthesis with a polymorphic type would require some additional setup:
\<close>
lemma plus_param_polymorphic: "(op+, op+)\<in>Id\<rightarrow>Id\<rightarrow>Id" by auto
schematic_goal 
  assumes [autoref_rules]: "(l,s)\<in>\<langle>Id\<rangle>ls.rel"
  notes [autoref_rules] = plus_param_polymorphic
  shows "(RETURN ?c,abs_sum s)\<in> \<langle>Id\<rangle>nres_rel"
  unfolding abs_sum_def[abs_def]
  apply (autoref_monadic (plain))
  done

text \<open>The correctness lemma is proven as before, 
  but the transitivity argument has fewer steps\<close>
lemma "(l,s)\<in>\<langle>Id\<rangle>ls.rel \<Longrightarrow> ls_sum_impl' l = \<Sum>s"
proof -
  assume A: "(l,s)\<in>\<langle>Id\<rangle>ls.rel"
  note ls_sum_impl'.refine[THEN nres_relD]
  also note abs_sum_correct[of s]
  finally show ?thesis using A 
    by (auto simp: br_def)
qed

text \<open>And for red-black trees\<close>

schematic_goal rs_sum_impl':
  assumes [autoref_rules]: "(l,s)\<in>\<langle>nat_rel\<rangle>rs.rel"
  shows "(RETURN ?c,abs_sum s)\<in> \<langle>nat_rel\<rangle>nres_rel"
  unfolding abs_sum_def[abs_def]
  apply (autoref_monadic (plain))
  done
concrete_definition rs_sum_impl' uses rs_sum_impl'
export_code rs_sum_impl' in SML

lemma "(l,s)\<in>\<langle>Id\<rangle>rs.rel \<Longrightarrow> rs_sum_impl' l = \<Sum>s"
proof -
  assume A: "(l,s)\<in>\<langle>Id\<rangle>rs.rel"
  note rs_sum_impl'.refine[THEN nres_relD]
  also note abs_sum_correct[of s]
  finally show ?thesis using A 
    by (auto simp: br_def)
qed



end

